/*
 * a and b initiates a bulk send TCP transfer to c.
 * additional isochronous traffic (VoIP-like) between a-c, b-c
 * The default TCP version in ns-3, TcpNewReno, is used as the transport-layer
 * protocol.
 */

#include <iostream>
#include <fstream>
#include <string>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/config-store-module.h"
#include "ns3/error-model.h"
#include "ns3/tcp-header.h"
#include "ns3/udp-header.h"
#include "ns3/enum.h"
#include "ns3/event-id.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/ipv4-flow-classifier.h"
#include "ns3/flow-monitor-helper.h"
#include "ns3/dsf-queue.h"

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("CoDelTest");

void
printStats (FlowMonitor::FlowStats st, FlowId id, Ipv4Address sourceIP, uint16_t sourcePort, Ipv4Address destIP, uint16_t destPort, uint64_t run, float simDuration, std::string desc)
{
  std::cout << desc << "," << run << "," << id << ",sIP," << sourceIP << std::endl;
  std::cout << desc << "," << run << "," << id << ",sPo," << sourcePort << std::endl;
  std::cout << desc << "," << run << "," << id << ",dIP," << destIP << std::endl;
  std::cout << desc << "," << run << "," << id << ",dPo," << destPort << std::endl;
  std::cout << desc << "," << run << "," << id << ",Dur," << simDuration << std::endl;
  std::cout << desc << "," << run << "," << id << ",TxB," << st.txBytes << std::endl;
  std::cout << desc << "," << run << "," << id << ",RxB," << st.rxBytes << std::endl;
  std::cout << desc << "," << run << "," << id << ",TxP," << st.txPackets << std::endl;
  std::cout << desc << "," << run << "," << id << ",RxP," << st.rxPackets << std::endl;
  std::cout << desc << "," << run << "," << id << ",LsP," << st.lostPackets << std::endl;
  if (st.rxPackets > 0)
    {
      std::cout << desc << "," << run << "," << id << ",MDl," << (st.delaySum.GetSeconds() / st.rxPackets) << std::endl;
    }
  for (uint32_t i=0; i<st.delayHistogram.GetNBins (); i++)
    std::cout << desc << "," << run << "," << id << "," << st.delayHistogram.GetBinStart (i) << "-"<< st.delayHistogram.GetBinEnd (i) << "," <<st.delayHistogram.GetBinCount (i) << std::endl;
} 

AddressValue
CreateTcpSink (uint16_t port, Ptr< Node > node, Ipv4Address ip, float stopTime)
{
  PacketSinkHelper sinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), port));
  ApplicationContainer sinkApp = sinkHelper.Install (node);
  sinkApp.Start (Seconds (0));
  sinkApp.Stop (Seconds (stopTime));
  AddressValue remoteAddress (InetSocketAddress (ip, port));
  return remoteAddress;
}

void
CreateBulkFlow (AddressValue remoteAddress, Ptr<Node> sender, uint32_t pktSize, float stopTime)
{
  Ptr<BulkSendApplication> app = CreateObject<BulkSendApplication>();
  app->SetAttribute ("Remote", remoteAddress);
  app->SetAttribute( "Protocol", StringValue("ns3::TcpSocketFactory") );
  app->SetAttribute( "MaxBytes", UintegerValue(0) );
  app->SetAttribute ("SendSize", UintegerValue (pktSize));
  app->SetAttribute( "StartTime", TimeValue(Seconds(0)) );
  app->SetAttribute( "StopTime", TimeValue(Seconds(stopTime - 3)) );
  app->SetTOS(1<<2);
  sender->AddApplication(app);
}

void
CreateOnOffFlow (AddressValue remoteAddress, Ptr<Node> sender, float stopTime)
{
  Ptr<OnOffApplication> app = CreateObject<OnOffApplication>();
  app->SetAttribute ("Remote", remoteAddress);
  app->SetAttribute( "Protocol", StringValue("ns3::TcpSocketFactory") );
  app->SetAttribute( "PacketSize", UintegerValue(280) );
  app->SetAttribute( "StartTime", TimeValue(Seconds(0)) );
  app->SetAttribute( "StopTime", TimeValue(Seconds(stopTime - 3)) );
  app->SetTOS(0<<2);
  sender->AddApplication(app);
}

void
installDSF (Ptr<PointToPointNetDevice> device, list<DSFClass> classList, double bottleneckRate)
{
  Ptr<DSFQueue> dsf = new DSFQueue();
  dsf->setLinkRate(bottleneckRate);
  for ( list<DSFClass>::iterator it = classList.begin(); it != classList.end(); ++ it ) {
    dsf->AddClass(*it);
  }
  device->SetQueue(dsf);
}

int main (int argc, char *argv[])
{
  std::string hostDelay = "5ms";
  std::string routerDelay = "1ms";
  std::string queueType = "CoDel";
  uint32_t pktSize = 1400;
  static const uint64_t mega = 1000000;
  double bottleneckRate = 10;
  double defaultRate = 100;

  uint64_t queueTime   = 200;        // queue time in ms
  uint32_t numOfDownLoadBulkFlows = 20;   // # of download bulk transfer flows
  uint32_t numOfDownLoadOnOffFlows = 1;   // # of download onoff flows
  uint64_t run = 1;

  std::string desc = queueType;

  float startTime = 0.1;
  float simDuration = 120;        //in seconds
  double delayBinWidth=0.001;
  uint32_t codelInterval = 100;      // The CoDel algorithm interval in ms
  uint32_t codelTarget = 5;          // The CoDel algorithm target queue delay in ms

  CommandLine cmd;
  cmd.AddValue ("queueType", "bottleneck queue type", queueType);
  cmd.AddValue ("queueTime",   "queue time in ms",    queueTime);
  cmd.AddValue ("numOfDownLoadBulkFlows", "number of download bulk transfer flows", numOfDownLoadBulkFlows);
  cmd.AddValue ("numOfDownLoadOnOffFlows", "number of download OnOff flows", numOfDownLoadOnOffFlows);
  cmd.AddValue ("codelInterval", "the CoDel algorithm interval in ms", codelInterval);
  cmd.AddValue ("codelTarget", "the CoDel algorithm target queue delay in ms", codelTarget);
  cmd.AddValue ("delayBinWidth", "width of delay width in report", delayBinWidth);
  cmd.AddValue ("startTime", "simulation start time", startTime);
  cmd.AddValue ("simDuration", "simulation duration in seconds", simDuration);
  cmd.AddValue ("run", "run number", run);
  cmd.AddValue ("desc", "description of run for output csv", desc);
  cmd.Parse (argc, argv);

  bottleneckRate = bottleneckRate * mega;
  defaultRate = defaultRate * mega;
  float stopTime = startTime + simDuration;
  RngSeedManager::SetRun(run);

  // Create the nodes
  NS_LOG_INFO ("Create nodes");
  NodeContainer nodes;
  nodes.Create (9);

  Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (pktSize));
  InternetStackHelper stack;
  stack.InstallAll ();

  // name network 
  Names::Add ("hosta", nodes.Get (0));
  Names::Add ("routera", nodes.Get (1));
  Names::Add ("hostb", nodes.Get (2));
  Names::Add ("routerb", nodes.Get (3));
  Names::Add ("routerc", nodes.Get (4));
  Names::Add ("hostc", nodes.Get (5));
  Names::Add ("hostd", nodes.Get (6));
  Names::Add ("hoste", nodes.Get (7));
  Names::Add ("hostf", nodes.Get (8));

  PointToPointHelper p2p;
  // bottlenecks links
  p2p.SetChannelAttribute ("Delay", StringValue (hostDelay));
  p2p.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(bottleneckRate)));
  if (queueType.compare ("DropTail") == 0 || queueType.compare ("DSF") == 0 )
    {
      p2p.SetQueue ("ns3::DropTailQueue",
                    "Mode", StringValue ("QUEUE_MODE_BYTES"),
                    "MaxBytes", UintegerValue ( (double(queueTime) / 8000) * bottleneckRate));
    }
  else if (queueType.compare ("CoDel") == 0)
    {
      p2p.SetQueue ("ns3::CoDelQueue",
                    "Mode", StringValue ("QUEUE_MODE_BYTES"),
                    "MaxBytes", UintegerValue ( (double(queueTime) / 8000) * bottleneckRate),
                    "Interval", TimeValue(MilliSeconds(codelInterval)),
                    "Target", TimeValue(MilliSeconds(codelTarget)));
    }

  NetDeviceContainer hostaLink = p2p.Install (NodeContainer (Names::Find<Node>("hosta"), Names::Find<Node>("routera")));
  NetDeviceContainer hostbLink = p2p.Install (NodeContainer (Names::Find<Node>("hostb"), Names::Find<Node>("routerb")));
  NetDeviceContainer hostcLink = p2p.Install (NodeContainer (Names::Find<Node>("routerc"), Names::Find<Node>("hostc")));
  if (queueType.compare ("DSF") == 0 )
    {
      std::cout << "dsf" << std::endl;
      list<DSFClass> classList;
      DSFClass c1;
      c1.num = 0;
      c1.delay = 10;
      classList.push_back(c1);
      DSFClass c2;
      c2.num = 1;
      c2.delay = 200;
      classList.push_back(c2);
      // install DSF queues on all bottleneck links
      // std::cout << "node" << std::endl;
      // Ptr<Node> n = Names::Find<Node>("hosta");
      // std::cout << "dev" << std::endl;
      // std::cout << n->GetDevice(0)<< std::endl;
      // std::cout << "install" << std::endl;
      installDSF(hostaLink.Get(0)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      installDSF(hostbLink.Get(0)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      installDSF(hostcLink.Get(0)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      installDSF(hostaLink.Get(1)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      installDSF(hostbLink.Get(1)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      installDSF(hostcLink.Get(1)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      // installDSF(Names::Find<Node>("hostb")->GetDevice(0)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);
      // installDSF(Names::Find<Node>("routerc")->GetDevice(0)->GetObject<PointToPointNetDevice>(), classList, bottleneckRate);

    }


  // no-bottleneck links
  p2p.SetChannelAttribute ("Delay", StringValue (routerDelay));
  p2p.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(defaultRate)));
  p2p.SetQueue ("ns3::DropTailQueue",
                "Mode", StringValue ("QUEUE_MODE_PACKETS"),
                "MaxPackets", UintegerValue (1000));
  NetDeviceContainer hostdLink = p2p.Install (NodeContainer (Names::Find<Node>("routera"), Names::Find<Node>("hostd")));
  NetDeviceContainer hosteLink = p2p.Install (NodeContainer (Names::Find<Node>("routerb"), Names::Find<Node>("hoste")));
  NetDeviceContainer hostfLink = p2p.Install (NodeContainer (Names::Find<Node>("hostf"), Names::Find<Node>("routerc")));
  NetDeviceContainer routeracLink = p2p.Install (NodeContainer (Names::Find<Node>("routera"), Names::Find<Node>("routerc")));
  NetDeviceContainer routerbcLink = p2p.Install (NodeContainer (Names::Find<Node>("routerb"), Names::Find<Node>("routerc")));

  // assign addresses
  // host links
  Ipv4AddressHelper ipv4;
  ipv4.SetBase ("10.1.1.0", "255.255.255.0");
  Ipv4InterfaceContainer hostaIPs = ipv4.Assign (hostaLink);
  ipv4.SetBase ("10.1.2.0", "255.255.255.0");
  Ipv4InterfaceContainer hostbIPs = ipv4.Assign (hostbLink);
  ipv4.SetBase ("10.1.3.0", "255.255.255.0");
  Ipv4InterfaceContainer hostcIPs = ipv4.Assign (hostcLink);
  ipv4.SetBase ("10.1.4.0", "255.255.255.0");
  Ipv4InterfaceContainer hostdIPs = ipv4.Assign (hostdLink);
  ipv4.SetBase ("10.1.5.0", "255.255.255.0");
  Ipv4InterfaceContainer hosteIPs = ipv4.Assign (hosteLink);
  ipv4.SetBase ("10.1.6.0", "255.255.255.0");
  Ipv4InterfaceContainer hostfIPs = ipv4.Assign (hostfLink);
  // inner routers
  ipv4.SetBase ("10.1.98.0", "255.255.255.0");
  ipv4.Assign (routeracLink);
  ipv4.SetBase ("10.1.99.0", "255.255.255.0");
  ipv4.Assign (routerbcLink);

  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

  /////// bg streams
  // a to d
  AddressValue remoteAddress = CreateTcpSink (10000, Names::Find<Node>("hostd"), hostdIPs.GetAddress (1), stopTime);
  uint32_t blkFlwsc=numOfDownLoadBulkFlows;
  while (blkFlwsc)
    {
      CreateBulkFlow (remoteAddress, Names::Find<Node>("hosta"), pktSize, stopTime);
      blkFlwsc--;
    }

  // b to e
  remoteAddress = CreateTcpSink (11000, Names::Find<Node>("hoste"), hosteIPs.GetAddress (1), stopTime);
  blkFlwsc=numOfDownLoadBulkFlows;
  while (blkFlwsc)
    {
      CreateBulkFlow (remoteAddress, Names::Find<Node>("hostb"), pktSize, stopTime);
      blkFlwsc--;
    }

  // f to c
  remoteAddress = CreateTcpSink (12000, Names::Find<Node>("hostc"), hostcIPs.GetAddress (1), stopTime);
  blkFlwsc=numOfDownLoadBulkFlows;
  while (blkFlwsc)
    {
      CreateBulkFlow (remoteAddress, Names::Find<Node>("hostf"), pktSize, stopTime);
      blkFlwsc--;
    }

  /////// monitored TCP streams
  // a to c
  remoteAddress = CreateTcpSink (30002, Names::Find<Node>("hostc"), hostcIPs.GetAddress (1), stopTime);
  blkFlwsc=numOfDownLoadBulkFlows;
  while (blkFlwsc)
    {
      CreateBulkFlow (remoteAddress, Names::Find<Node>("hosta"), pktSize, stopTime);
      blkFlwsc--;
    }

  // b to c
  remoteAddress = CreateTcpSink (40002, Names::Find<Node>("hostc"), hostcIPs.GetAddress (1), stopTime);
  blkFlwsc=numOfDownLoadBulkFlows;
  while (blkFlwsc)
    {
      CreateBulkFlow (remoteAddress, Names::Find<Node>("hostb"), pktSize, stopTime);
      blkFlwsc--;
    }

  /////// monitored (VOIP-like) streams  
  // a to c
  remoteAddress = CreateTcpSink (30001, Names::Find<Node>("hostc"), hostcIPs.GetAddress (1), stopTime);
  uint32_t ooFlc=numOfDownLoadOnOffFlows;
  while (ooFlc)
    {
      CreateOnOffFlow (remoteAddress, Names::Find<Node>("hosta"), stopTime);
      ooFlc--;
    }


  // b to c
  remoteAddress = CreateTcpSink (40001, Names::Find<Node>("hostc"), hostcIPs.GetAddress (1), stopTime);
  ooFlc=numOfDownLoadOnOffFlows;
  while (ooFlc)
    {
      CreateOnOffFlow (remoteAddress, Names::Find<Node>("hostb"), stopTime);
      ooFlc--;
    }

  // Flow Monitor
  FlowMonitorHelper flowmonHelper;
  Ptr<FlowMonitor> monitor; 
  monitor = flowmonHelper.InstallAll ();
  monitor->SetAttribute("DelayBinWidth", DoubleValue(delayBinWidth));

  Simulator::Stop (Seconds (stopTime));
  Simulator::Run ();

  monitor->CheckForLostPackets ();
  std::map< FlowId, FlowMonitor::FlowStats > stats = monitor->GetFlowStats();
  Ptr<Ipv4FlowClassifier> classifier = DynamicCast<Ipv4FlowClassifier> (flowmonHelper.GetClassifier ());
  for (std::map<FlowId, FlowMonitor::FlowStats>::const_iterator i = stats.begin (); i != stats.end (); ++i)
    {
      Ipv4FlowClassifier::FiveTuple t = classifier->FindFlow (i->first);
      if (t.destinationPort==40001 || t.destinationPort==40002 || t.destinationPort==30001 || t.destinationPort==30002)
	{
	  printStats(i->second, i->first, t.sourceAddress, t.sourcePort, t.destinationAddress, t.destinationPort, run, simDuration, desc);
	}
    }

  Simulator::Destroy ();
  return 0;
}
